import functools
import typing
import locale
from . import utility
from . import translate


class I18next:
    def __init__(self, **options):
        self.translation_map = {}
        self._options = {}
        self._plural_func = None
        self.update_options(**options)
        if self._options.get('init_immediate', True) and self._options.get('backend'):
            self.load()

    def update_options(self,
                       resources=None,
                       backend=None, 
                       init_immediate: bool = True,
                       lng: typing.Union[str, list, tuple]=None,
                       ns: typing.Union[str, list, tuple]=None,
                       default_lng=None,
                       default_ns=None,
                       fallback_lng: typing.Union[str, list, tuple]=None,
                       fallback_ns: typing.Union[str, list, tuple]=None,
                       plural_rules=None,
                       return_objects=False,
                       join_arrays=None,
                       **options):
        self._options.update({
            'resources': resources,
            'backend': backend,
            'init_immmediate': init_immediate,
            'lng': lng,
            'ns': ns,
            'default_lng': default_lng,
            'default_ns': default_ns,
            'fallback_lng': fallback_lng,
            'fallback_ns': fallback_ns,
            'plural_rules': plural_rules,
            'return_objects': return_objects,
            'join_arrays': join_arrays,
            **options})
        if self._options.get('plural_rules'):
            self._plural_func = utility.get_plural_func(self._options['plural_rules'])
        if self._options.get('resources'):
            self.translation_map = self._options.get('resources')
        self.translate = translate.Translate(self)

    def load(self):
        all_lngs = utility.merge_list(self._options.get('lng'),
                                      self._options.get('default_lng'), 
                                      self._options.get('fallback_lng'))
        all_ns = utility.merge_list(self._options.get('ns'),
                                    self._options.get('default_ns'),
                                    self._options.get('fallback_ns'))
        if backend := self._options.get('backend'):
            self.translation_map = backend.read_all(all_lngs, all_ns)

    def resolve_lng_ns(self, key_info, **kwargs):
        lng = kwargs.get('lng') or self._options.get('default_lng') or locale.getlocale()[0]
        ns = kwargs.get('ns') or key_info.get('ns') or self._options.get('default_ns')
        return (lng, ns)
    
    def resolve_translation(self, lng, ns, **kwargs):
        fallback_lng = self._options.get('fallback_lng') or []
        fallback_ns = self._options.get('fallback_ns') or []
        translation = None
        for curr_lng in utility.merge_list(lng, fallback_lng):
            for curr_ns in utility.merge_list(ns, fallback_ns):
                translation = utility.get_deep_value(self.translation_map, [curr_lng, curr_ns])
                if translation is not None:
                    break
        return translation
    
    def get_key_info(self, original_key, **kwargs):
        ret = {}
        key = original_key
        ns_splitted = key.split(':', maxsplit=1)
        # handle namespace
        if len(ns_splitted) == 2:
            ret['ns'] = ns_splitted[0]
            key = ns_splitted[1]
        if kwargs.get('context'):
            key = f'{key}_{kwargs["context"]}'
        if kwargs.get('count') and self._plural_func:
            key = f'{key}_{self._plural_func(kwargs["count"])}'
            print(key)
        # handle deep key
        ret['deep'] = key.split('.')
        ret['key'] = key
        return ret

    def t(self, key, *subs, **kwargs):
        ret = None
        key_info = self.get_key_info(key, **kwargs)
        (lng, ns) = self.resolve_lng_ns(key_info, **kwargs)
        translation = self.resolve_translation(lng, ns)
        if translation:
            ret = self.translate.translate_value(
                utility.get_deep_value(translation, key_info['deep']),
                **kwargs)
        if ret is None and subs:
            ret = subs[0]
        return ret

    def get_translate_func(self, **kwargs):
        return functools.partial(self.t, **kwargs)


'''
This module contains the file format loaders for the pyi18next package.
'''


def load_data(file, file_type: str) -> dict:
    '''
    Load the data from the file.
    '''
    content = file.read() if hasattr(file, 'read') else file
    if file_type == 'json':
        import json
        return json.loads(content)
    elif file_type == 'yaml' or file_type == 'yml':
        import yaml
        return yaml.safe_load(content)
    else:
        raise ValueError(f'Unsupported file type: {file_type}')


import os.path
import typing
from .. import loaders


class  Backend:
    def __init__(self, name_mapping=None):
        '''
        Initialize the backend with the name mapping function.
        name_mapping: a function that maps the namespace to the file name.
        '''
        self.set_name_mapping(name_mapping)

    def set_name_mapping(self, name_mapping):
        if isinstance(name_mapping, dict):
            self.name_mapping = lambda lng, ns: name_mapping.get(ns, ns)
        elif callable(name_mapping):
            self.name_mapping = name_mapping
        else:
            self.name_mapping = lambda lng, ns: ''        

    def read_one(self, lng: str, ns: str):
        '''
        Read the translation data from the file.
        '''
        filename = self.name_mapping(lng, ns)
        file_name, file_ext = os.path.splitext(filename)
        file_type = file_ext[1:].lower()
        with open(filename, 'r') as f:
            data = loaders.load_data(f, file_type)
        return data
    
    def read_all(self, lng_list: typing.List, ns_list: typing.List):
        '''
        Read all the translation data from the files.
        '''
        ret = {}
        for lng in lng_list:
            ret.setdefault(lng, {})
            for ns in ns_list:
                ret[lng][ns] = self.read_one(lng, ns)
        return ret


import re
import typing
import babel.plural

def merge_list(*args):
    '''
    Merge the lists into one list.
    '''
    ret = []
    for arg in args:
        if not arg:
            continue
        elif isinstance(arg, (list, tuple)):
            ret.extend(arg)
        else:
            ret.append(arg)
    return ret


def match_lng(lng1: str, lng2: str, ignore_case: bool=True, implicit: bool=False):
    '''
    Check if the two language codes match.
    '''
    if ignore_case:
        lng1, lng2 = [i.lower() for i in [lng1, lng2]]
    if not implicit:
        return lng1 == lng2
    else:
        short_lng, long_lng = sorted([lng1, lng2], key=len)
        return long_lng.startswith(short_lng)


def get_deep_value(obj, keys, default=None):
    '''
    Get a value from obj based on a chain of keys.
    '''
    value = default
    curr = obj
    for key in keys:
        if isinstance(curr, dict) and key in curr:
            value = curr[key]
            curr = value
        elif isinstance(curr, (list, tuple)) and (isinstance(key, int) or key.isdigit()):
            value = curr[int(key)]
            curr = value
        else:
            return default
    return value

def get_plural_func(rules: typing.Union[str, dict]):
    '''
    Get the plural rules object from the rules string.
    '''
    rule_func = None
    rule_dict = {}
    if isinstance(rules, str):
        for rule in rules.split(';'):
            if not rule.strip():
                continue
            key, value = rule.split(':')
            key = key.strip()
            value = value.strip()
            rule_dict[key] = value
    elif isinstance(rules, dict):
        rule_dict = rules
    if rule_dict:
        rule_func = babel.plural.PluralRule(rule_dict)
    return rule_func



import re
import json


class Translate:
    def __init__(self, i18next):
        self._i18next = i18next
        self._set_patterns()

    def _set_patterns(self):
        quote = r'(?:\'|\")'
        itpl_delim_beg = re.escape('{{')
        itpl_delim_end = re.escape('}}')
        nest_delim_beg = re.escape('$t(')
        nest_delim_end = re.escape(')')
        comma_delim = r'(?:\s*,\s*)'  # comma with optional spaces
        json_pattern = r'(?P<json>\{.*?\})?'
        param_name = r'(?:\s*;?\s*[\w]+)'
        param_format = r'(?: (?:[\d]+) | (?:\'.+\') | (?:[\d]+\.[\d]*))'
        param_default = fr'(?:\s*:\s*{param_format})?'
        func_pattern = fr'(?:(?:\w+)\((?P<params>(?:{param_name}{param_default})*)\))'
        param_pairs = fr'(?:{param_name}{param_default})'
        func_pattern_named = fr'(?:(?P<func_name>\w+)\((?P<params>(?:{param_name}{param_default})*)\))'
        func_list_pattern = fr'(?P<func_list>(?:{comma_delim}{func_pattern})*)'
        identifier = r'[\w\.]+'
        self.pattern_params = re.compile(param_pairs, re.VERBOSE)
        self.pattern_func = re.compile(func_pattern_named, re.VERBOSE)
        self.pattern = re.compile(fr"""
        (?:
          {itpl_delim_beg}(?P<itpl>{identifier}){func_list_pattern}{itpl_delim_end}
        ) |
        (?:
          {nest_delim_beg}(?P<nest>{identifier})(?:{comma_delim}{json_pattern})?{nest_delim_end}
        )
        """, re.VERBOSE)

    def parse_params(self, params):
        ret = {}
        for param_mo in self.pattern_params.finditer(params):
            ret[param_mo.group('param_name')] = param_mo.group('param_default')
        return ret
    
    def parse_func(self, func_list):
        ret = []
        for func_mo in self.pattern_func.finditer(func_list):
            ret.append(
                (func_mo.group('func_name'), self.parse_params(func_mo.group('params')))
            )
        return ret

    def handle_itpl(self, mo, **kwargs):
        identifier = mo.group('itpl')
        func_list = mo.group('func_list')
        if func_list:
            func_list = self.parse_func(func_list)
        value = kwargs.get(identifier)
        return value
    
    def handle_nest(self, mo, **kwargs):
        identifier = mo.group('nest')
        json_data = mo.group('json')
        if json_data:
             json_data = json.loads(json_data)
        else:
            json_data = {}
        return self._i18next.t(identifier, **json_data)

    def translate_str(self, value, **kwargs):
        def _sub_replace(mo):
            if mo.group('itpl'):
                return self.handle_itpl(mo, **kwargs)
            elif mo.group('nest'):
                return self.handle_nest(mo, **kwargs)
            raise ValueError(f"Invalid match object: {mo}")
        ret = self.pattern.sub(_sub_replace, value)
        return ret
    
    def translate_object(self, value, **kwargs):
        ret = {}
        for k, v in value.items():
            ret[k] = self.translate_value(v, **kwargs)
        if kwargs.get('return_objects') or self._i18next._options.get('return_objects'):
            return ret
        else:
            return None
    
    def translate_array(self, value, **kwargs):
        ret = []
        join_arrays = kwargs.pop('joinArrays', None) or self._i18next._options.get('join_arrays')
        for v in value:
            ret.append(self.translate_value(v, **kwargs))
        if join_arrays is not None:
            return join_arrays.join(ret)
        elif kwargs.get('return_objects') or self._i18next._options.get('return_objects'):
            return ret
        else:
            return None
    
    def translate_value(self, value, **kwargs):
        if isinstance(value, str):
            return self.translate_str(value, **kwargs)
        elif isinstance(value, dict):
            return self.translate_object(value, **kwargs)
        elif isinstance(value, list):
            return self.translate_array(value, **kwargs)
        else:
            return value

